/***************************************************************************
* Copyright (c) 2016, Johan Mabille and Sylvain Corlay                     *
*                                                                          *
* Distributed under the terms of the BSD 3-Clause License.                 *
*                                                                          *
* The full license is in the file LICENSE, distributed with this software. *
****************************************************************************/

#ifndef XPROPERTY_CONFIG_HPP
#define XPROPERTY_CONFIG_HPP

#define XPROPERTY_VERSION_MAJOR 0
#define XPROPERTY_VERSION_MINOR 10
#define XPROPERTY_VERSION_PATCH 3

#endif
